package de.mutschler.rpdl.entity;

import de.mutschler.rpdl.basicvalue.*;
import de.mutschler.rpdl.*;
import java.util.*;

/**
 * This class represents the access to an object in the
 * tree.
 */
public class FormulaObject extends RpdlFormula implements java.io.Serializable
{
	static final long serialVersionUID = 2725956894512328866L;
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------

	/** The name of the object */
	private String objName;

	/** This is a directly access object. */
	private final static int DIRECT = 0;

	/** This object means the parent object. */
	private final static int BASE = 1;
	private final static String STR_BASE = "base";

	/** This object means the parent object. */
	private final static int THIS = 2;
	private final static String STR_THIS = "this";
	
	/** This object means the parent object. */
	private final static int PARENT = 3;
	private final static String STR_PARENT = "parent";
	
	/** This object means all parent object. */
	private final static int PARENTALL = 4;
	private final static String STR_PARENTALL = "parentall";
	
	/** the type of object. */
	private int type;

	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------
	
	/** This objects points to the current object. */
	public FormulaObject (String name) {
		objName = name;
		if (STR_PARENT.equalsIgnoreCase(name)) {
			type = PARENT;
		} 
		else if (STR_BASE.equalsIgnoreCase(name)) {
			type = BASE;
		} 
		else if (STR_THIS.equalsIgnoreCase(name)) {
			type = THIS;
		} 
		else if (STR_PARENTALL.equalsIgnoreCase(name)) {
			type = PARENTALL;
		}
		else {
			type = DIRECT;
		}
	}

	//----------------------------------------------------------------------
	//------- Methods
	//----------------------------------------------------------------------

	/** Get the name obj the object. */
	public String getName() {return objName;}
	
	/** Get the actual object, where this link points to.
	 * @param obj the object to start navigation.
	 */
	public RpdlObject getObject(RpdlBase main, RpdlObject obj) throws RpdlException {
		RpdlObject res;
		switch (type) {
		case THIS:		res = obj; break;
		case DIRECT:	res = main.getObject(objName); break;
		case BASE:		res = main.getBaseObject(); break;
		case PARENT:
			Iterator it = obj.parents().iterator();
			if (!it.hasNext()) throw new RpdlException("Object " + obj.getName() + " has no parent object!");
			res = (RpdlObject)it.next();	// just take the first parent
			break;
		case PARENTALL:
			throw new RpdlException("PARENTALL Not supported in this method.");
		default:
			throw new RpdlException("unknown type for objectaccess: " + type);
		}
		if (res == null) {
			throw new RpdlException("Didn't get a valid object for '" + objName + "' type=" + type);
		}
		return res;
	}

	/** Use this if you can handle multiple objects, such as parentall. */
	public Set getObjects(RpdlBase main, RpdlObject obj) throws RpdlException {
		Set res = new HashSet();
		if (type == PARENTALL) {
			// add all the parent objects of the current object
			for (Iterator it=obj.parents().iterator(); it.hasNext();) {
				res.add(it.next());
			}
			return res;
		}
		else {
			res.add(getObject(main, obj));
		}
		return res;
	}
	
	//----------------------------------------------------------------------
	//------- Methods for interface RpdlFormula
	//----------------------------------------------------------------------
	
	/**
	 * Perform a calculation, and return the result.
	 */
	public BasicValue calc(RpdlBase main, RpdlObject obj) throws RpdlException {
		return new ValueObject(getObject(main, obj));
	}

	/**
	 * Get the type of result.
	 */
	public String getResultType(RpdlBase main, Entity e) {
		return ValueObject.TYPENAME;
	}

	/**
	 * Get a list of all dependencies for this formula.
	 * This returns a Set of RpdlObjects.
	 */
	public Set getDependencies(RpdlBase main, RpdlObject obj) throws RpdlException {
		Set s = new HashSet();
		s.add(getObject(main, obj) );
		return s;
	}

	public String toString() {
		return "($" + objName + ")";
	}
}