package de.mutschler.rpdl.entity;

import de.mutschler.rpdl.*;
import de.mutschler.rpdl.basicvalue.*;
import de.mutschler.util.*;
import de.mutschler.rpdl.calc.*;
import java.util.*;

/**
 * for faster access to the name, this class copies the name from the attributedefinition
 * to the local class.
 */
public class RpdlAttribute extends NamedObj
{
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------
	/** The Definition for this attribute. */
	private RpdlAttributeDef attrDef;
	
	/** The value of this attribute */
	private BasicValue value;
	
	/** The associated object. */
	private RpdlObject object;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------
	/**
	 * This constructor uses the name of the definition
	 */
	public RpdlAttribute(RpdlAttributeDef ad, RpdlObject obj) throws RpdlException {
		super(ad.getName());
		attrDef = ad;
		object = obj;
		value =  BasicValue.createValue(ad.getType());
    }

	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------
	
	/** Get the attribute definition. */
	public RpdlAttributeDef getAttributeDef() {return attrDef;}
	
	/** Set the value of this attribute. Note: this value is calculated from the formulas
	 * in the calcList. */
	public void setValue(BasicValue v) throws RpdlConversionException
	{
		value.setValue(v);
	}
	
	/** Get the value of this attribute. */
	public BasicValue getValue() {return value;}

	/** Get a string representation of the value. */
	public final String getStringValue() {return value.getStringValue();}
	
	/** Get the type of this attribute. */
	public String getType() {return attrDef.getType();}
	
	/** Get the associated object. */
	public RpdlObject getObject() {return object;}
	
	/** Get the RpdlBase from the associated object. */
	private RpdlBase getMain() {return object.getMain();}
	
	//----------------------------------------------------------------------
	//------- calc List handling
	//----------------------------------------------------------------------
	
	/** List of Formulas for the calculation of this attribute.
	 * Items in this list are of type <code>StmtHandle</code>. */
//	private List calcList = new ArrayList();
	private SortedSet calcList = new TreeSet();
	
	/** Get the statements used for calculating this attribute. */
	public Iterator calcStatements() {return calcList.iterator();}
	
	/**
	 * display the contents of the calc list.
	 */
	String calcListString() {return listString(calcList.iterator());}

	/** Add a StmtHandle to the calcList. If the handle is already in the
	 *  List, an exception is thrown.
	 */
	public void addCalcHandle(StmtHandle sh) throws RpdlException {
		if (calcList.contains(sh)) {
/*			StmtHandle sh2=null;
			for (Iterator it = calcList.iterator(); it.hasNext();) {
				sh2 = (StmtHandle) it.next();
				if (sh2.equals(sh)) break;
				sh2 = null;
			}
			if (sh2 == null) {
				error("Method TreeSet.contains() is buggy. It found statement '" + sh + "' although " + 
				"it is not in the list:");
				debug("calcList of " + toString() +":");
				for (Iterator it = calcList.iterator(); it.hasNext();) {
					debug(it.next().toString());
				}
			}
			else
*/				throw new RpdlException("Handle " + sh + " already in calcList!");
		}
/*		// get the dependencies of the statement
		Set depSet = sh.getStatement().getDependencies(getMain(), sh.getObject());
		// and check if this attribute is inside.
		if (!depSet.contains(this)) {
			// we have don't have a local cycle
			debug("resetting calclist for attr '" + toString() + "' for statement '" + sh +"'");
			calcList.clear();
		}
*/		calcList.add(sh);
		debug("added " + sh + " to calcList of attr " + getName());
		// add to the dependencies:
		addDependencies(sh);
		// add notification event
		getMain().addCalcItem(new CalcAttr(this));
	}
	
	/** Remove a calc handle. */
	public void removeCalcHandle(StmtHandle sh) throws RpdlException {
		if (!calcList.contains(sh))
			throw new RpdlException("Handle '" + sh + "' doesn't exist in calcList of attribute '$" + getObject().getName() + "." + getName() + "'");
		calcList.remove(sh);
		// add notification event
		getMain().addCalcItem(new CalcAttr(this));
	}
	/**
	 * Calculate the value of the attribute. Therefore all the statements in
	 * the calcList are executed.
	 */
	public void calculate(RpdlBase main) throws RpdlException {
		// check if we have statements in the calclist. if not,
		// then throw an exception!
		if (calcList.size() == 0) {
			throw new RpdlException("Attribute '$" + getObject().getName() + "." + getName() + "' is not initialized!", getObject().getOrigin());
		}
		// for performance reason, only trigger the attributes
		// in the triggerlist if the value has changed. Therefore store the current
		// value.
		BasicValue oldValue = (BasicValue)value.clone();
		for (Iterator it=calcList.iterator(); it.hasNext(); ) {
			StmtHandle sh = (StmtHandle)it.next();
			sh.execute(main);
		}
		// now, compare the values:
		if (value == oldValue ||
			((ValueBool)value.calculate(BasicValue.COMP_EQ, oldValue)).getValue()) {
			debug("Value of attribute " + toString() + " hasn't changed. Not triggering further attributes.");
			return;
		}
		// now trigger all the statements:
		for (Iterator it = triggerStatements(); it.hasNext(); ) {
			StmtHandle sh = (StmtHandle)it.next();
			debug("triggering handle " + sh + " for recalculation.");
			sh.trigger(main);
		}
		
	}
	
	//--------- dependencies

	/* Here a set of attributes which this formula depends on is stored.
	 * This is used to check for cycles, and create the calculation order.
	 */
	 
	/** The set of current dependencies. Set this to null, to perform
	 *  a recalculation.
	 */
	private Set dependencies = null;
	
	/** Get a set of dependencies. They are all RpdlAttributes. */
	public Set getDependencies() throws RpdlException {
		if (dependencies == null) calculateDependencies();
		return dependencies;
	}
	
	/** Scan all the dependencies from the calc-list. */
	private void calculateDependencies() throws RpdlException {
		for (Iterator it=calcList.iterator(); it.hasNext(); ) {
			StmtHandle sh = (StmtHandle)it.next();
			addDependencies(sh);
		}
		// for the case we don't have any dependencies...
		if (dependencies == null) dependencies = new HashSet();
	}
	
	/** Add the dependencies of the statement of the handle to the set.
	 * This function will add the dependencies to all attributes
	 * of the triggerlist too. So here we can perform the cycle check
	 * for the calculation:
	 * if the current attribute is in the set to be added, then we have a cycle.
	 */
	private void addDependencies(StmtHandle sh) throws RpdlException {
		debug("Adding dependencies from handle " + sh + " to attribute " + toString());
		try {
			Set newDep = sh.getStatement().getDependencies(getMain(), sh.getObject());
			newDep.remove(this);	// remove the dependency from myself. To prevent a cycle on my own.
			addDependencies(newDep);
		} catch (RpdlException re) {
			re.fillOrigin(sh.getStatement().getOrigin());
		}
	}


	private void displaySet(Set s) {
		debug("displaying contents of set #" + s.size() + ":");
		int idx=1;
		for (Iterator it=s.iterator(); it.hasNext();) {
			debug(CR + "#" + idx + ": " + it.next());
		}
	}

	/** Well, this is the actual method to propagate the dependencies,
	 * This is just for performance reason, to prevent the creation
	 * of the set from the handle.
	 */
	protected void addDependencies(Set dep) throws RpdlException {
		ddebug("new dependencies for " + toString());
		//displaySet(dep);
//		try {
			if (dep.contains(this)) {
				throw new RpdlException(
					"Cycle in calculation detected: " + CR +
					"$" + getObject().getName() + "." + getName() + " [" + getOrigin() + "]"
					);
			}
			// now add the dependencies to the current set
			if (dependencies == null) dependencies = new HashSet();
			if (dep.size() == 0) {
				debug("no dependencies to set.");
				return;
			}
			dependencies.addAll(dep);
			// now propagte the new dependencies to the target.
			// Note: the current element is not added, since that is already set.
			for (Iterator it = targets(); it.hasNext();) {
				RpdlAttribute target = (RpdlAttribute)it.next();
				debug("Propagating dependencies to attribute " + target.toString());
				target.addDependencies(dep);
			}
/*		} catch (RpdlException re) {
			throw new RpdlException(
				re.getMessage() + CR + 
				"$" + getObject().getName() + "." + getName() + " [" + getOrigin() + "]"
				);
		} /**/
	}

	/**
	 * Check, if the this attribute must be calculated before the
	 * given attribute.
	 */
	public final boolean calcBefore(RpdlAttribute attr) throws RpdlException{
		return attr.getDependencies().contains(this);
	}

	//----------------------------------------------------------------------
	//------- trigger List handling
	//----------------------------------------------------------------------

	/** List of Attributes to trigger, when this attribute changes.
	 * Items in this list are of type <code>StmtHandle</code>. */
	private List triggerList = new ArrayList();
	
	/** formatted triggerList. */
	String triggerListString() {return listString(triggerStatements());}

	/** Get all the StmtHandles, which will be triggered by this attribute. */
	public Iterator triggerStatements() {return triggerList.iterator();}

	/** Add a StmtHandle to the calcList. If the handle is already in the
	 *  List, an exception is thrown.
	 */
	public void addTriggerHandle(StmtHandle sh) throws RpdlException {
		if (triggerList.contains(sh))
			throw new RpdlException("Handle " + sh + " already in triggerList!");
		ddebug("added " + sh+ " to triggerList of attr " + getName());
		triggerList.add(sh);
		// add all dependencies to the targets:
		for (Iterator it = sh.getStatement().getTargets(getMain(), sh.getObject()).iterator(); it.hasNext();) {
			RpdlAttribute target = (RpdlAttribute)it.next();
			target.addDependencies(getDependencies());
		}
	}
	
	/** Get all the target attributes. This does not return the own attribute! */
	public Iterator targets() throws RpdlException {
		return new Iterator() {
			Iterator handles = triggerList.iterator();
			Iterator stmttargets = null;
			RpdlAttribute nextAttr;
			
			// constructor
			{
				setNext();
			}
			
			public boolean hasNext() {return nextAttr != null;}
			/** Set the next valid RpdlAttribute */
			private void setNext() {
				nextAttr = getNext();
				while(nextAttr != null && nextAttr.equals(this))
					nextAttr = getNext();
			}

			// read the next attribute
			private RpdlAttribute getNext() {
				try{
					if (stmttargets == null) {
						if (!handles.hasNext()) {
							// no elements in the triggerlist
							return null;
						}
						StmtHandle sh = (StmtHandle)handles.next();
						stmttargets = sh.getStatement().getTargets(getMain(), sh.getObject()).iterator();
					}
					while (stmttargets.hasNext() || handles.hasNext()) {
						// do we have anything in the statements, then return true
						if (stmttargets.hasNext()) {
							return (RpdlAttribute)stmttargets.next();
						}
						// advance to the next handle
						StmtHandle sh = (StmtHandle)handles.next();
						stmttargets = sh.getStatement().getTargets(getMain(), sh.getObject()).iterator();
					}
				} catch(RpdlException re) {
					Log.exception(re);
				}
				return null;
			}
			
			// get the next target; this is a RpdlAttribute
			public Object next() {
				Object o = nextAttr;
				setNext();
				return o;
			}
			
			public void remove() {throw new UnsupportedOperationException("cannot remove!");}
		};
	}

	// misc
	/** create a formatted output of a list of stmthandles. */
	private String listString(Iterator it) {
		StringBuffer sb = new StringBuffer(300);
		boolean first=true;
		int idx=1;
		while(it.hasNext()) {
			if (!first) sb.append(CR);
			first = false;
			sb.append("#").append(idx).append(": ");
			sb.append(it.next().toString());
			idx++;
		}
		return sb.toString();
	}
	
	//----------------------------------------------------------------------
	//------- Overridden system methods
	//----------------------------------------------------------------------
	
	public boolean equals(Object o) {
		if (o instanceof RpdlAttribute) {
			RpdlAttribute attr = (RpdlAttribute)o;
			return	getName().equals(attr.getName()) &&
					getObject().equals(attr.getObject());
		}
		return false;
	}
	
	public String toString() {
		return "$" + getObject().getName() + "." + getName() + "(=" + value.getStringValue()+ ")";
	}
	
}