package de.mutschler.rpdl.entity;

import de.mutschler.rpdl.*;
import de.mutschler.rpdl.statement.*;
import de.mutschler.util.TreeFormatter;
import de.mutschler.util.StringUtil;
import de.mutschler.rpdl.calc.CalcAttr;
import java.util.*;

public class RpdlObject extends Entity
{
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	/** Create a new RpdlObject with the corresponding name. */
	public RpdlObject(RpdlBase rbase, String s, Set types) {
		super(rbase, s, types);
		
	}

	/**
	 * Create a plain copy from another object. This does not
	 * copy the links, parents etc.  Only the parent types and statements
	 * are copied.
	 */
	public RpdlObject(Entity orig, RpdlBase base, String s) throws RpdlException {
		super(orig, base, s);
	}
	//----------------------------------------------------------------------
	//------- History
	//----------------------------------------------------------------------
	
	/** A link to the current history for this object. If this object is in
	 * the object-library there is no history.
	 */
	private RpdlHistoryEntry historyEntry = null;

	/** Set the History. */	
	public void setHistoryEntry(RpdlHistoryEntry he) {historyEntry = he;}
	/** Get the History. */	
	public RpdlHistoryEntry getHistoryEntry() {return historyEntry;}
	
	//----------------------------------------------------------------------
	//------- Subobjects
	//----------------------------------------------------------------------
	
	/** A List of subobjects. Note: They are in a specific order! */
	private List subObjects = new ArrayList();
	
	/** Get an ordered list of subobjects */
	public Iterator getSubObjects() {return subObjects.iterator();}
	
	/**
	 * Get an ordered list of subobjects. Only Subobjects of the
	 * give type are returned. 
	 */
	public Iterator getSubObjects(RpdlType type) {
		// if null is passed, return all subobjects
		if (type == null) return getSubObjects();
		// create a new list, and put there all objects of the given type
		List l = new ArrayList(subObjects.size());
		for (Iterator it=getSubObjects(); it.hasNext();) {
			RpdlObject ro = (RpdlObject)it.next();
			if (ro.isType(type)) {
				l.add(ro);
			}
		}
		// finally return the iterator of the newly created list
		return l.iterator();
	}
	
	/** Get a specific subobject. If the subobject cannot be found, null is returned. */
	public RpdlObject getSubObject(String name) {
		for (Iterator i=subObjects.iterator(); i.hasNext(); ) {
			RpdlObject ro = (RpdlObject)i.next();
			if (name.equals(ro.getName()))
				return ro;
		}
		return null;
	}
	
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------
	
	/** List of Attributes. Contains mapping (String)name -> (RpdlAttribute)attribute. */
	private Map attributes = new HashMap();
	
	/** add this attribute to the list.
	 * @throws RpdlException if the attribute already exists
	 */
	public void addAttribute(RpdlAttribute attr) throws RpdlException {
		if (attributes.containsKey(attr.getName()))
			throw new RpdlException("Attribute '" + attr.getName() + "' already exists in object '" + getName() + "'");
		attributes.put(attr.getName(), attr);
	}
	/** add this attribute to the list. Note, this uses its own name.
	 * This method is mainly for the table, to have their own name.
	 */
	void addAttribute(String name, RpdlAttribute attr) {attributes.put(attr.getName(), attr);}
	/** add this attribute to the list.
	 * @throws RpdlException if the attribute already exists
	 */
	public void removeAttribute(String attrName) throws RpdlException {
		RpdlAttribute attr = getAttribute(attrName);
		if (attr == null) {
			throw new RpdlException("Attribute '" + attrName + "' cannot be removed from object '" + getName() + "' because it doesn't exist.");
		}
		// check the notification lists:
		StringBuffer err = null;
		for (Iterator it=attr.triggerStatements(); it.hasNext();) {
			if (err == null)  {
				err = new StringBuffer(100);
				err.append("Attribute '" + attrName + "' on Object '" + getName() + "' can't be removed. The following statements depend on it:");
			}
			
			StmtHandle sh = (StmtHandle)it.next();
			err.append(CR);
			err.append("depend: ");
			err.append(sh.toString());
		}
		for (Iterator it=attr.calcStatements(); it.hasNext();) {
			if (err == null)  {
				err = new StringBuffer(100);
				err.append("Attribute '" + attrName + "' on Object '" + getName() + "' can't be removed. The following statements depend on it:");
			}
			
			StmtHandle sh = (StmtHandle)it.next();
			err.append(CR);
			err.append("calc: ");
			err.append(sh.toString());
		}
		if (err != null)
			throw new RpdlException(err.toString());
		attributes.remove(attrName);
		// remove queued calculation for this attribute
		getMain().removeCalcItem(new CalcAttr(attr));
	}
	/** Get an attribute. returns null, if the attribute isn't found. */
	public RpdlAttribute getAttribute(String name) {return (RpdlAttribute)attributes.get(name);}

	/** for debugging. */
	public String getAttributesString() {
		StringBuffer sb = new StringBuffer(200);
		boolean start = true;
		int i=1;
		for (Iterator it = attributes.values().iterator(); it.hasNext(); ) {
			RpdlAttribute attr = (RpdlAttribute)it.next();
			if (!start) sb.append(CR);
			sb.append("Attr #").append(StringUtil.fmt(i++,2)).append(": ");
			sb.append("(").append(attr.getType()).append(") ");
			sb.append(attr.getName() + " = '" + attr.getValue().getStringValue() + "'");
			sb.append(CR + "calclist:" + CR);
			sb.append(StringUtil.indent(attr.calcListString()));
			sb.append(CR + "triggerlist:" + CR);
			sb.append(StringUtil.indent(attr.triggerListString()));
			start = false;
		}
		return sb.toString();
	}

	//----------------------------------------------------------------------
	//------- Parent handling
	//----------------------------------------------------------------------

	/** A collection of parent objects. Mapping is (String)name -> (RpdlObject)parentobject. */
	private Set parentObjects = new HashSet();
	
	/** Add this Object to the parent Object. */
	void addParent(RpdlObject parent) throws RpdlException {
		if (parentObjects.contains(parent))
			throw new RpdlException("Internal Error: Object '" + getName() + "' already has parent Object '" + parent.getName() + "'!");
		parentObjects.add(parent);
	}
	
	/** Remove a parent Object. */
	void removeParent(RpdlObject parent) throws RpdlException {
		if (!parentObjects.contains(parent))
			throw new RpdlException("Internal Error: Object '" + getName() + "' doesn't have parent Object '" + parent.getName() + "'!");
		parentObjects.remove(parent);
	}
	
	/** Get an iterator for the parent-objects. */
	public Set parents() {
		return parentObjects;
	}

	/**
	 * put the parents in the form
	 * + sub1
	 * | + sub11
	 * | + sub12
	 * + sub2
	 *   + sub21
	 */
	public String parentsString() {
		TreeFormatter tf = new TreeFormatter() {
			public String getInfo(Object o)		{return ((RpdlObject)o).getName();}
			public Iterator subitems(Object o)	{return ((RpdlObject)o).parents().iterator();}
		};
		return tf.getSubTreeString(this);
    }
	
	//----------------------------------------------------------------------
	//------- Link/Unlink
	//----------------------------------------------------------------------

	/** link an object to this object as a subobject. */
	public void link (RpdlObject subObj) throws RpdlException {
		if (getSubObject(subObj.getName()) != null)
			throw new RpdlException("Object '" + subObj.getName() + "' already has a link to '" + getName() + "'!");
		
		debug("adding object '" + subObj.getName() + "' to '" + getName() + "'");
		
		// check for cycle
		try {
			checkLinkCycle(subObj, this);
		} catch (RpdlException re) {
			throw new RpdlException(
				"Cycle detected while trying to link " + subObj.getName() + " to " + getName() + ":" + CR +
				"Path: " + re.getMessage());
		}
		
		// add the object to this one
		subObjects.add(subObj);
		// add this object as parent to the subobject
		subObj.addParent(this);
		// update the current TreeWalker
		getMain().getObjectOrder().treeChanged();
		// check the base-state
		subObj.checkBase();
	}
	
	/** Method to check, if linking this object would create a cycle in the
	 *  tree. If so, an exception is thrown.
	 * If there exists a path from start to end, throw the exception.
	 * The path taken through the subobjects.
	 */
	private void checkLinkCycle(RpdlObject start, RpdlObject end) throws RpdlException {
		debug("checking cycle for " + start.getName() + " -> " + end.getName());
		if (start.getName().equals(end.getName())) {
			throw new RpdlException(end.getName() + " [" + end.getOrigin() + "]");
		}

		try {
			for (Iterator it = start.getSubObjects(); it.hasNext(); ) {
				RpdlObject ro = (RpdlObject) it.next();
				checkLinkCycle(ro, end);
			}
		} catch (RpdlException re) {
			throw new RpdlException(re.getMessage() + " -> " + start.getName() + " [" + start.getOrigin() + "]");
		}
	}
	
	/** unlink and object. */
	public void unlink(RpdlObject subObj) throws RpdlException {
		if (getSubObject(subObj.getName()) == null)
			throw new RpdlException("Object '" + subObj.getName() + "' isn't linked to '" + getName() + "'!");
			
		debug("removing object '" + subObj.getName() + "' from '" + getName() + "'");

		// remove the object from the subobject list
		subObjects.remove(subObj);
		// remove this one from the parent list of the subobject
		subObj.removeParent(this);
		// update the current TreeWalker
		getMain().getObjectOrder().treeChanged();
		// check the base-state
		subObj.checkBase();
	}
	
	//----------------------------------------------------------------------
	//------- Statement handling
	//----------------------------------------------------------------------
	/** Add a statement to the internal lists.
	 * overridden from the superclass to automatically parse the statement
	 * when the object is in the base.
	 */
	public void addStatement(Stmt s) throws RpdlException {
		super.addStatement(s);
		if (inBase()) {
			addHandle(new StmtHandle(this, s));
		}
	}
	
	/** List of StatementHandles. */
	private List stmtHandles = null;
	
	/** Create all the statement-handles, and initialze them. */
	private void createStmtHandles() throws RpdlException {
		if (stmtHandles != null) throw new RpdlException ("internal: statements already initialized!");
		debug("Initializing handles for object " + getName());
		stmtHandles = new LinkedList();
		typesLoaded = new HashSet();
		// scan through the parent types (recursively)
		for (Iterator it = parentTypes(); it.hasNext(); ) {
			RpdlType pt = (RpdlType)it.next();
			addTypeStmts(pt);
		}
		// now the own statements
		for (Iterator it = getStatements(); it.hasNext(); ) {
			Stmt s = (Stmt)it.next();
			debug("adding stmt '" + s + "' to object " + getName());
			addHandle(new StmtHandle(this, s));
		}
	}

	/** Add a StmtHanndle to the list, and perform initialization. */
	private void addHandle(StmtHandle sh) {
		debug("Adding Handle " + sh);
		// check for duplicate statements
		StmtHandle same = findHandle(sh);
		if (same != null) {
			debug("Statement '" + sh.getStatement() + "' already in the statement list of object " + getName());
			return;
		}
		// perform the initialization of the statement
		try {
			sh.onBaseLink(getMain(), true);
			// after successful init, add it to the list
			stmtHandles.add(sh);
		} catch (RpdlException re) {
			error("Failed to initialize statement '" + sh + "' [" + sh.getStatement().getOrigin() +"]", re);
		}
	}
	
	/** Find a handle with the same statement. returns null if no handle is found. */
	private StmtHandle findHandle(StmtHandle sh) {
		Stmt s = sh.getStatement();
		for (Iterator it = stmtHandles.iterator(); it.hasNext(); ) {
			StmtHandle sht = (StmtHandle)it.next();
			if (s.equals(sht.getStatement()))
				return sht;
		}
		return null;
	}
	
	/**
	 * This is a container to store all the types we have loaded.
	 * This is for cycle-detection in the parent type tree.
	 */
	private Set typesLoaded;
	/**
	 * Add the statements of this type to the stmt-handle list.
	 * This is done recusively.
	 */
	private void addTypeStmts(RpdlType type) throws RpdlException {
		if (typesLoaded.contains(type.getName())) {
			warn("Type " + type.getName() + " for object " + getName() + " already loaded. skipping!");
			return;
		}
		// add the parent types
		for (Iterator it = type.parentTypes(); it.hasNext(); ) {
			RpdlType pt = (RpdlType)it.next();
			addTypeStmts(pt);
		}
		
		// add the statements:
		for (Iterator it = type.getStatements(); it.hasNext(); ) {
			Stmt s = (Stmt)it.next();
			StmtHandle sh = new StmtHandle(this, s);
			addHandle(sh);
		}
		
	}
	
	/**
	 * Remove all the handles. This method goes through the
	 * list of StmtHandles and calls onBaseUnlink().
	 */
	private void removeStmtHandles() throws RpdlException {
		if (stmtHandles == null) throw new RpdlException ("internal: statements not initialized!");
		debug("Removing handles for object " + getName());
		// go through the list of handles, and remove them. 
		// this has to be done in reverse order, for a good cleanup
		// sort the handles in reverse order:
		Object handles[] = stmtHandles.toArray();
		Arrays.sort(handles);
		for (int idx=handles.length-1; idx>=0; idx--) {
			StmtHandle sh = (StmtHandle)handles[idx];
			try {
				sh.onBaseLink(getMain(), false);
			} catch (RpdlException re) {
				error("Failed to remove statement '" + sh + "' [" + sh.getStatement().getOrigin() +"]", re);
			}
		}
		stmtHandles = null;	// kill the list.
	}

	/** Get the List as a string for displaying. */
	public String getStmtListString() {
		if (stmtHandles == null)
			return "No Statements initialized yet";
		
		StringBuffer sb = new StringBuffer(200);
		boolean start = true;
		int i=1;
		for (Iterator it = stmtHandles.iterator(); it.hasNext(); ) {
			StmtHandle sh = (StmtHandle)it.next();
			if (!start) sb.append(CR);
			sb.append("Stmt #").append(StringUtil.fmt(i++,2)).append(": ");
			sb.append(sh.toString());
			start = false;
		}
		return sb.toString();
	}
	//----------------------------------------------------------------------
	//------- Base handling
	//----------------------------------------------------------------------
	
	/** If this attribute has the base set, then to object is linked to the base. */
	private RpdlObject base = null;
	
	/** get the Base Object. */
	public RpdlObject getBaseObject() {return base;}
	
	/** Method to check, if the object is in the base-tree */
	public boolean inBase() {return base != null;}
	/**
	 * Main Methods to perform the adding/remvoing to the base object.
	 * It checks the state of the parent class, and if there are
	 * differences, then the appropiate work will be done.
	 */
	public void checkBase() throws RpdlException {
		boolean parentbase = false;	
		RpdlObject baseObj = null;
		for (Iterator it = parents().iterator(); it.hasNext(); ) {
			RpdlObject p = (RpdlObject) it.next();
			if (p.inBase()) {
				parentbase = true;
				baseObj = p.getBaseObject();
				break;
			}
		}
		// if both are equal, nothing has to be done.
		if (parentbase == inBase()) return;
        
		if (inBase()) {
			// No parent is in the base, so we have been removed.
			removeBase();
		}
		else {
			setBaseObject(baseObj);
		}
	}

	/**
	 * This object is added to the base-tree. So perform the necessary setup.
	 */
	public void setBaseObject(RpdlObject base) throws RpdlException {
		this.base = base;
		debug ("adding object " + getName() + " to the base tree.");
//		if (!inBase()) throw new RpdlException("Object " + getName() + " is already in the tree!");
		/* Create the instances of all the statements.
		 * Therefore get all the statements from the parent-types!
		 * Then the statements of this object.
		 */
		createStmtHandles();
		getMain().triggerObjectChange(RpdlBase.NOTIFY_OBJECT_CHANGE, this);
		
	}

	/**
	 * This object is not in the base-tree anymore. So remove any listeners, etc.
	 * the links have already been removed.
	 */
	private void removeBase() throws RpdlException {
		debug ("removing object " + getName() + " from the base tree.");
		// remove all my handles
		removeStmtHandles();
		// notify about the changes
		getMain().triggerObjectChange(RpdlBase.NOTIFY_OBJECT_CHANGE, this);
		// at least, remove the base
		this.base = null;
	}

	/** Add another type to the list. If the type already exists,
	 * nothing is done.
	 * @return if the new type has not been there before.
	 */
	public boolean addType(RpdlType type) throws RpdlException  {
		boolean res = super.addType(type);
		if (res) return true;
		// if we are inside the base object, perform the necessary updates
		if (inBase()) {
			addTypeStmts(type);
		}
		return res;
	}
	/** Objects are equal if their name is equal. */
	public boolean equals(Object o) {
		if (o instanceof RpdlObject) {
			return getName().equals(((RpdlObject)o).getName());
		}
		return false;
	}

	public String toString() {
		return
			"OBJECT  " + super.toString() +
			CR + "ParentObjects:" + CR + parentsString() + CR +
			CR + "active Statements:" + CR + getStmtListString() + CR +
			CR + "Attributes:" + CR + getAttributesString() + 
			CR;
	}
}
