package de.mutschler.rpdl.statement;

import de.mutschler.util.LogObj;
import de.mutschler.rpdl.*;
import de.mutschler.rpdl.entity.*;
import de.mutschler.rpdl.calc.*;

import java.util.*;


/**
 * This is the superclass of all statements.
 * a statement is always bind to an entity where it comes from.
 * This is independent from the object(s) using the statement!
 */
public abstract class Stmt extends LogObj implements java.io.Serializable
{
	static final long serialVersionUID = -1372009172724002209L;
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------

	/** The Entity which this statement comes from.
	 * The statement can be in multiple objects, but this statement
	 * comes from the same parenttype.
	 */
	private Entity derivedEntity=null;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	public Stmt() {
		// get a unique ID for the statement
		ID = getNextID();
	}	

	//----------------------------------------------------------------------
	//------- ID handling
	//----------------------------------------------------------------------
	/** the unique ID for a statement. */
	private int ID;
	/** identify the statement by its ID. */
	public int getID() {return ID;}
	
	// create a unique ID. */
	private static int nextID = 0;
	private synchronized static int getNextID() {
		return nextID++;
	}
	
	/** If the IDs are the same, we have the same statement. */
	public final boolean equals(Object o) {
		return getID() == ((Stmt)o).getID();
	}
	
	public void setDerivedEntity(Entity e) throws RpdlException {
		if (derivedEntity != null)
			throw new RpdlException("Cannot set the derived entity of a statement multiple times");
		derivedEntity = e;
	}
	
	//----------------------------------------------------------------------
	//------- calculation
	//----------------------------------------------------------------------
	
	/**
	 * This method can be overwritten, and then used to check the validity
	 * of the statement. This method is called from the addCommand() of CmdCrType & CmdCrObj
	 * The default implementaion does nothing.
	 */
	public void checkStatement(RpdlBase main, Entity e) throws RpdlException {}
	
	/**
	 * This method is called, when an object is linked/unlinked to the base object.
	 * called from StmtHandle.
	 * @param link set this to true, if the handle is linked, and to false, if this handle is unlinked.
	 */
	public void onBaseLink(RpdlBase main, RpdlObject obj, StmtHandle sh, boolean link) throws RpdlException {
		// now add the handle to the calcList of the resulting attribute
		Set targets = getTargets(main, obj);
		Set depSet = getDependencies(main, obj);

		// handle the dependencies: add this statement to the triggerlist
		// of all depending attributes
		for (Iterator it=depSet.iterator(); it.hasNext(); ) {
			Object dep = it.next();
			if (dep instanceof RpdlAttribute) {
				RpdlAttribute attr = (RpdlAttribute) dep;
				// do not add the attribute if it is in both lists
				// This allows local cycles.
				if (targets.contains(attr)) {
//					isCycle = true;	// set the cycle-flag
					continue;
				}

				if (link) {
					// statement is linked
					attr.addTriggerHandle(sh);
				}
				else {
					// statement is unlinked
					//attr.removeTriggerHandle(sh);
				}
				continue;
			}
			error("unknown dependency " + dep + " in statement '" + toString() + "' at " +  getOrigin());
		}
		
		// handle the targets
		// add this handle to the calcList of every Attribute of the dependencies of the formula
		for (Iterator it=targets.iterator(); it.hasNext(); ) {
			Object dest = it.next();
			debug("setting calc for " + dest);
			if (dest instanceof RpdlAttribute) {
				RpdlAttribute attr = (RpdlAttribute) dest;
				if (link) {
					// statement is linked
					attr.addCalcHandle(sh);
				}
				else {
					// statement is unlinked
					attr.removeCalcHandle(sh);
				}
				continue;
			}
			error("unknown target " + dest + " in statement '" + toString() + "' at " +  getOrigin());
		}
	}
	
	/**
	 * Excecute the statement. Here the actual calculation for SET-statements
	 * are done. 
	 * called from StmtHandle.
	 */
	public void execute(RpdlBase main, RpdlObject obj) throws RpdlException {
		debug("executing statement '" + toString() + "' on object " + obj.getName());
	}

	/**
	 * This Method is called, when a statement is triggered. You can add
	 * things to the CalcHandler in this method.
	 *
	 * Note: All the target attributes are added to the calcHandler in the
	 * Stmt class.
	 */
	public void trigger(RpdlBase main, StmtHandle sh) throws RpdlException {
		for (Iterator it = getTargets(main, sh.getObject()).iterator(); it.hasNext();) {
			Object o = it.next();
			if (o instanceof RpdlAttribute) {
				RpdlAttribute attr = (RpdlAttribute)o;
				debug("triggering attribute " + attr + " for recalculation.");
				main.addCalcItem(new CalcAttr(attr));
			}
		}
	}

	//----------------------------------------------------------------------
	//------- targets & dependencies
	//----------------------------------------------------------------------
	
	/**
	 * This returns a set of dependencies for this statement. These are of type
	 * <ul>
	 * <li><b>RpdlObject</b>: if the statement depends on an object</li>
	 * <li><b>RpdlAttribute</b>: if the statement depends on an attribute</li>
	 * </ul>
	 */
	public Set getDependencies(RpdlBase main, RpdlObject obj) throws RpdlException {
		return new HashSet();
	}
	
	/**
	 * Get all the target attributes this statement will modify.
	 */
	public Set getTargets(RpdlBase main, RpdlObject obj) throws RpdlException {
		return new HashSet();
	}
	
	//----------------------------------------------------------------------
	//------- sorting
	//----------------------------------------------------------------------
	
	/** If the statement should not be executed. This is the default. */
	public static final int PRI_NONE = -1;
	/** Statement should be executed at the high priority. */
	public static final int PRI_HIGH = 3;
	// Note pri norm must have a difference of 2 to the next pri,
	// to allow local modifications. see Stmt_Set.
	/** Statement should be executed at the normal priority. (e.g. SET) */
	public static final int PRI_NORM = 6;
	/** Statement should be executed after the others. e.g. ASSERT. */
	public static final int PRI_LOW = 10;
	
	/** Get the execution priority. */
	public int getPriority() {
		throw new RuntimeException("no pri set.!");
//		return PRI_NONE;
	}
	
	//-------------------
	// Origin handling
	private Origin origin = null;
	/** Get the origin of this command. */
	public Origin getOrigin() {return origin;}
	/** Set the origin of this command. */
	public void setOrigin(Origin o) {origin=o;}
	
}