package de.mutschler.rpdl.statement;

//import de.mutschler.util.LogObj;
import de.mutschler.rpdl.*;
import de.mutschler.rpdl.entity.*;
import de.mutschler.rpdl.calc.*;
import de.mutschler.rpdl.basicvalue.*;
import java.util.*;

public class Stmt_Attr extends Stmt_Set implements java.io.Serializable
{
	static final long serialVersionUID = 4897551680961931068L;
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------

	/** The name of the attribute */
	private String name;

	/** Flag if this attribute is required to be initialized.*/
	private boolean required;
	
	/** Initial value of the attribute. This can be null! */
//	private BasicValue init;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	public Stmt_Attr(String name, boolean required, RpdlFormula init) {
		super(new FormulaAttribute(name), init, BasicValue.COMP_EQ);
		this.name = name;
		this.required = required;
//		this.init = init;
	}
	
	/** init is Formula with empty string. */
	public Stmt_Attr(String name, boolean required) {
		this(name, required, null);
	}
	
	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------
	
	/** Get the name for this attribute. */
	public String getName() {return name;}
	/** check if this attribute requires initialization. */
	public boolean isRequired() {return required;}

	//----------------------------------------------------------------------
	//------- Overridden methods from Stmt
	//----------------------------------------------------------------------
	
	/** Get the execution priority. */
	public int getPriority() {return PRI_HIGH;}

	/**
	 * This method can be overwritten, and then used to check the validity
	 * of the statement. This method is called from the addCommand() of CmdCrType & CmdCrObj
	 * The default implementaion does nothing.
	 */
	public void checkStatement(RpdlBase main, Entity e) throws RpdlException {
		// check if the attribute exists
		RpdlAttributeDef attrDef = main.getAttributeDef(name);
		if (attrDef == null) {
			throw new RpdlException("Attribute '" + name + "' not defined.", getOrigin());
		}
		// if this attribute has the required flag set, then don't
		// create an inital formula
		if (isRequired()) return;
		// not required
		if (getFormula() == null) {
//			RpdlAttributeDef attrDef = main.getAttributeDef(name);
			debug("setting inital formula for attribute " + name);
			setFormula(new FormulaConst(BasicValue.createValue(attrDef.getType())));
		}
		super.checkStatement(main, e);
    }
	/**
	 * This method is called, when an object is linked/unlinked to the base object.
	 */
	public void onBaseLink(RpdlBase main, RpdlObject obj, StmtHandle sh, boolean link) throws RpdlException {
		if (link) {
			debug("creating attr '" + getName() + "' for object '" + obj.getName() + "'");
			RpdlAttributeDef attrDef = main.getAttributeDef(name);
			if (attrDef == null)
				throw new RpdlException("Attribute '" + getName() + " not defined.", getOrigin());
			RpdlAttribute attr = new RpdlAttribute(attrDef, obj);
			attr.setOrigin(getOrigin());
			obj.addAttribute(attr);
			/* force recalculation:
			 * This is for the case, the attribute has the flag required, and
			 * is not initialized. With this statement, the recalculation wouldn't happen
			 * and we can only recognize the missing initialization on runtime-usage.
			 */
			main.addCalcItem(new CalcAttr(attr));
			// in order to handle the initial assignment, simply call the superclass
			try {
				super.onBaseLink(main, obj, sh, link);
			} catch (Exception e) {
				error ("while linking attr: " + toString());
					de.mutschler.util.Log.exception(e);
			}
		}	// link
		else {
			// unlink
			debug("removing attr '" + getName() + "' from object '" + obj.getName() + "'");
			// first remove the statement from the calculations
			super.onBaseLink(main, obj, sh, link);
			// then remove the actual attribute
			obj.removeAttribute(name);
		}
	}

	// The dependencies: note: if the flag required is set, then
	// empty Sets are returned

	/**
	 * Get all the target attributes this statement will modify.
	 * This is an empty set, if the required flag is set.
	 */
	public Set getTargets(RpdlBase main, RpdlObject obj) throws RpdlException {
		if (isRequired()) return Collections.EMPTY_SET;
		else			  return super.getTargets(main, obj);
	}
	
	/**
	 * This returns a set of dependencies for this statement.
	 * This is an empty set, if the required flag is set.
	 */
	public Set getDependencies(RpdlBase main, RpdlObject obj) throws RpdlException {
		if (isRequired()) return Collections.EMPTY_SET;
		else			  return super.getDependencies(main, obj);
	}
	
	/**
	 * Excecute the statement.
	 * Do not execute if the required flag is set.
	 */
	public void execute(RpdlBase main, RpdlObject obj) throws RpdlException {
		if (isRequired()) return;
		super.execute(main, obj);
	}
	//----------------------------------------------------------------------
	//------- Overridden system methods
	//----------------------------------------------------------------------
	
	public String toString() {
		return "(" + getID() + ") " + (isRequired()?"REQUIRED ":"") + "ATTRIBUTE " + getName() + " = " + getFormula() + ";";
	}
}
