package de.mutschler.util;

import java.util.*;
import java.text.*;

/**
 * This class represents a multi language string. 
 */
public class LString extends LogObj implements java.io.Serializable
{
	static final long serialVersionUID = 1875426849295582350L;
	/**
	 * The current locale. Note this is a static field. So the language is
	 * the same for all instances.
	 */
	private static Locale currentLocale = Locale.getDefault();
	
	/** Get the current locale. */
	public static Locale getLocale() {return currentLocale;}	
	
	/** Set the current locale. */
	public static void setLocale(Locale l) {currentLocale = l;}

	//--------------------------------------------------------------
	/**
	 * In this map the actual strings are stored. The mapping is
	 * Locale -> String
	 */
	private Map strings = new HashMap();
	
	public LString(LString ls) {
		strings.putAll(ls.getMap());
	}
	
	public LString() {}

	/**
	 * This cunstructs a new LString from a string.
	 * @see #parseString
	 */
	public LString(String s) {
		parseString(s);
	}
	
	/**	
	 * Get the map of the actual strings. The mapping is
	 * String (2-letter iso language) -> String
	 */
	public Map getMap() {return strings;}
	
	
	/**
	 * <p>The method parses a string of the form
	 * <li>abc {l1} def {l2} xyz</li>
	 * where lx specifies the 2-digit language and the following is the
	 * localized string in that language. If a language is missing at the
	 * beginning, the default is used. If you want to display a '{' or '}'
	 * in the text, then escape is with '\'</p>
	 * <p>e.g. "{de}\{in\} deutsch{en}in english" </p>
	 * <p>Note: The strings are added to the existing ones. </p>
	 */
	public void parseString(String in) throws IllegalArgumentException {
		Locale l = getLocale();
		StringBuffer strsb = new StringBuffer(20);	// store the actual string
		StringBuffer locsb = new StringBuffer(2);	// store the 2 char language
		// flag if we are reading a locale or the actual string
		boolean readLocale = false;
		// flag if an escape was previously read
		boolean escape = false;
		
		sloop: for (int idx=0; idx<in.length(); idx++) {
			char c = in.charAt(idx);
			if (escape) {
				escape = false;
				switch (c) {
				case 'n': (readLocale?locsb:strsb).append(CR); continue sloop;
				case '{':
				case '}':
					(readLocale?locsb:strsb).append(c); continue sloop;
				case '\\':	// I have "\\" -> save one, and set escape=true
				default:	// default is, save the escape char, and the continue parsing the next char
					(readLocale?locsb:strsb).append('\\'); break;
				}
					
			}
			switch(c) {
				case '{': {
					// when reading a { while in locale, then there is an error
					if (readLocale)
						throw new IllegalArgumentException("Duplicate '{' in lstring \"" + in + "\" at pos " + idx);
					// we are starting a new locale, so save the old string, if there is one
					ddebug("adding string '" + strsb + "' for locale " + l);
					strings.put(l.getLanguage(), strsb.toString());
					// reset the strings; starting a new localized string
					locsb.setLength(0);
					readLocale = true;	// start the locale
					break;
				}
				case '}': {
					// when reading a { while in locale, then there is an error
					if (!readLocale)
						throw new IllegalArgumentException("Found '}' without '{'  at pos " + idx + " in lstring \"" + in + "\"");
					// now check if the length of the locale string is ok
					if (locsb.length() != 2) 
						throw new IllegalArgumentException("The locale \"" + locsb + "\" must have a length of 2 characters at pos " + idx + " in lstring \"" + in + "\"");
					String iso2loc = locsb.toString().toLowerCase();
					l = new Locale(iso2loc, iso2loc);
					ddebug("new locale: " + l);
					// reset the strings; starting a new localized string
					strsb.setLength(0);
					readLocale = false;	// start the actual string
					break;
				}
				case '\\':
					escape = true;
					break;
				default:
					(readLocale?locsb:strsb).append(c);
					break;
			}
		}
		// save the last string

		// when reading a { while in locale, then there is an error
		if (readLocale)
			throw new IllegalArgumentException("Missing '}' in lstring \"" + in + "\"");
		// we are starting a new locale, so save the old string, if there is one
		ddebug("adding string '" + strsb + "' for locale " + l);
		strings.put(l.getLanguage(), strsb.toString());
		ddebug("result is: " + encodeString());
	}
	
	/** Simply calls encodeString(null). */
	public final String encodeString() {return encodeString(null);}

	/**
	 * This will encode the given string, but the string with the given language
	 * will be the default, and at the beginning of the resulting string.
	 * <p>example:<br>
	 * {de}deutsch{en}english
	 * encodeString(Locale.EN) -> english{de}deutsch
	 * @param defLocale if null, all strings will get thier langauge in front.
	 */
	public String encodeString(Locale defLocale) {
		StringBuffer sb = new StringBuffer(100);
		String defLang = (defLocale==null?null:defLocale.getLanguage());
		if (defLang != null) {
			String ls = getLocaleString(defLang);
			if (ls != null) {
				addEncStr(sb, ls);
			}
		}
		// handle the default language
		for (Iterator it = strings.keySet().iterator(); it.hasNext();) {
			String lang = (String)it.next();
			// skip the default language
			if (defLang != null && defLang.equals(lang))
				continue;

			String ls = getLocaleString(lang);
			sb.append('{').append(lang).append('}');
			addEncStr(sb, ls);
		}
		return sb.toString();
	}
	private void addEncStr(StringBuffer sb, String s) {
		// now take care of the '{' and '}' in the string
		for (int i=0; i<s.length(); i++) {
			char c = s.charAt(i);
			switch(c) {
			case '{':
			case '}':
				sb.append('\\');
			default:
				sb.append(c);
			}
		}
	}
	
	
	/** Get the default locale string. */
	public String getLocaleString() {return getLocaleString(getLocale());}
	/** Get a locale string. */
	public String getLocaleString(Locale l) {return getLocaleString(l.getLanguage());}
	private String getLocaleString(String lang) {
		Object res = strings.get(lang);
		if (res == null) {
			// not found: try the default language
			res = strings.get(getLocale().getLanguage());
		}
		if (res == null) {
			// get the first available string
			Iterator it = strings.values().iterator();
			if (it.hasNext())
				res = it.next();
		}
		if (res == null) {
			//error("no substring found in '" + encodeString() + "'");
			throw new RuntimeException("no substring found in '" + encodeString() + "'");
			//return "";	// if really nothing is found, return an empty string
		}
		return (String)res;
	}

	/** add/set a new localized string. This will overwrite a previously existing string. */
	public void setLocaleString(Locale l, String s) {strings.put(l.getLanguage(), s);}
	
	/** for easier handling, the toString displays the current locale string. */
	public String toString() {return getLocaleString();}
	
	
	//---------------------------------------------------------
	// testing

	public static void main(String[] args) {
		try {
			String test = "default{en}english";
			LString ls = new LString(test);
			out("test = " + test);
			out("getLocaleString()=" + ls.getLocaleString());
			out("getLocaleString(<en>)=" + ls.getLocaleString(Locale.ENGLISH));
			out("encodeString()=" + ls.encodeString());
			out("encodeString(<en>)=" + ls.encodeString(Locale.ENGLISH));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	private static void out(String s) {System.out.println(s);}
}